# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Specific modelling for "instantiated" linear expressions,
with only variables and literal coefficients.
"""
import dataclasses
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, TypeVar, Union

from gems.expression.expression import (
    OneScenarioIndex,
    ScenarioIndex,
    TimeIndex,
    TimeShift,
    TimeStep,
)

T = TypeVar("T")

EPS = 10 ** (-16)


def is_close_abs(value: float, other_value: float, eps: float) -> bool:
    return abs(value - other_value) < eps


def is_zero(value: float) -> bool:
    return is_close_abs(value, 0, EPS)


def is_one(value: float) -> bool:
    return is_close_abs(value, 1, EPS)


def is_minus_one(value: float) -> bool:
    return is_close_abs(value, -1, EPS)


@dataclass(frozen=True)
class TimeExpansion:
    """
    Carries knowledge of which timesteps this term refers to.
    Simplest one is "only the current timestep"
    """

    def get_timesteps(self, current_timestep: int, block_length: int) -> List[int]:
        return [current_timestep]

    def apply(self, other: "TimeExpansion") -> "TimeExpansion":
        """
        Apply another time expansion on this one.
        For example, a shift of -1 applied to a shift one +1 could provide
        a no-op TimeExpansion. Not yet supported for now, though.
        """
        return other


@dataclass(frozen=True)
class AllTimeExpansion(TimeExpansion):
    def get_timesteps(self, current_timestep: int, block_length: int) -> List[int]:
        return [t for t in range(block_length)]

    def apply(self, other: "TimeExpansion") -> "TimeExpansion":
        raise ValueError("No time operation allowed on all-time sum.")


@dataclass(frozen=True)
class TimeEvalExpansion(TimeExpansion):
    timestep: int

    def get_timesteps(self, current_timestep: int, block_length: int) -> List[int]:
        return [self.timestep]

    def apply(self, other: "TimeExpansion") -> "TimeExpansion":
        raise ValueError(
            "Time operation on evaluated expression not supported for now."
        )


@dataclass(frozen=True)
class TimeShiftExpansion(TimeExpansion):
    shift: int

    def get_timesteps(self, current_timestep: int, block_length: int) -> List[int]:
        return [current_timestep + self.shift]

    def apply(self, other: "TimeExpansion") -> "TimeExpansion":
        raise ValueError("Time operation on shifted expression not supported for now.")


@dataclass(frozen=True)
class TimeSumExpansion(TimeExpansion):
    from_shift: int
    to_shift: int

    def get_timesteps(self, current_timestep: int, block_length: int) -> List[int]:
        return [
            t
            for t in range(
                current_timestep + self.from_shift, current_timestep + self.to_shift
            )
        ]

    def apply(self, other: "TimeExpansion") -> "TimeExpansion":
        raise ValueError("Time operation on time-sums not supported for now.")


@dataclass(frozen=True)
class TermKey:

    """
    Utility class to provide key for a term that contains all term information except coefficient
    """

    component_id: str
    variable_name: str
    time_index: Optional[int]
    scenario_index: Optional[int]


def _str_for_coeff(coeff: float) -> str:
    if is_one(coeff):
        return "+"
    elif is_minus_one(coeff):
        return "-"
    else:
        return "{:+g}".format(coeff)


def _time_index_to_str(time_index: TimeIndex) -> str:
    if isinstance(time_index, TimeShift):
        if time_index.timeshift == 0:
            return "t"
        elif time_index.timeshift > 0:
            return f"t + {time_index.timeshift}"
        else:
            return f"t - {-time_index.timeshift}"
    if isinstance(time_index, TimeStep):
        return f"{time_index.timestep}"
    return ""


def _scenario_index_to_str(scenario_index: ScenarioIndex) -> str:
    if isinstance(scenario_index, OneScenarioIndex):
        return f"{scenario_index.scenario}"
    return ""


def _str_for_time_expansion(exp: TimeExpansion) -> str:
    if isinstance(exp, TimeShiftExpansion):
        return f".shift({exp.shift})"
    elif isinstance(exp, TimeSumExpansion):
        return f".sum({exp.from_shift}, {exp.to_shift})"
    elif isinstance(exp, AllTimeExpansion):
        return ".sum()"
    else:
        return ""


@dataclass(frozen=True)
class Term:
    """
    One term in a linear expression: for example the "10x" par in "10x + 5y + 5"

    Args:
        coefficient: the coefficient for that term, for example "10" in "10x"
        variable_name: the name of the variable, for example "x" in "10x"
    """

    coefficient: float
    component_id: str
    variable_name: str
    time_index: Optional[int]
    scenario_index: Optional[int]

    # TODO: It may be useful to define __add__, __sub__, etc on terms, which should return a linear expression ?

    def is_zero(self) -> bool:
        return is_zero(self.coefficient)

    def __str__(self) -> str:
        # Useful for debugging tests
        return repr(self)

    def __repr__(self) -> str:
        # Useful for debugging tests
        result = (
            f"{_str_for_coeff(self.coefficient)}{self.component_id}.{self.variable_name}"
            f"[{self.time_index},{self.scenario_index}]"
        )
        return result


def generate_key(term: Term) -> TermKey:
    return TermKey(
        term.component_id,
        term.variable_name,
        term.time_index,
        term.scenario_index,
    )


def _merge_dicts(
    lhs: Dict[TermKey, Term],
    rhs: Dict[TermKey, Term],
    merge_func: Callable[[Optional[Term], Optional[Term]], Term],
) -> Dict[TermKey, Term]:
    res = {}
    for k, left in lhs.items():
        right = rhs.get(k, None)
        res[k] = merge_func(left, right)
    for k, right in rhs.items():
        if k not in lhs:
            res[k] = merge_func(None, right)
    return res


def _add_terms(lhs: Optional[Term], rhs: Optional[Term]) -> Term:
    if lhs is not None and rhs is not None:
        return dataclasses.replace(rhs, coefficient=lhs.coefficient + rhs.coefficient)
    elif lhs is not None and rhs is None:
        return lhs
    elif lhs is None and rhs is not None:
        return rhs
    raise ValueError("Cannot add 2 null terms.")


def _substract_terms(lhs: Optional[Term], rhs: Optional[Term]) -> Term:
    if lhs is not None and rhs is not None:
        return dataclasses.replace(lhs, coefficient=lhs.coefficient - rhs.coefficient)
    elif lhs is not None and rhs is None:
        return lhs
    elif lhs is None and rhs is not None:
        return dataclasses.replace(rhs, coefficient=-rhs.coefficient)
    raise ValueError("Cannot subtract 2 null terms.")


class LinearExpression:
    """
    Represents a linear expression with respect to variable names, for example 10x + 5y + 2.

    Operators may be used for construction.

    Args:
        terms: the list of variable terms, for example 10x and 5y in "10x + 5y + 2".
        constant: the constant term, for example 2 in "10x + 5y + 2"

    Examples:
        Operators may be used for construction:

        >>> LinearExpression([], 10) + LinearExpression([Term(10, "x")], 0)
        LinearExpression([Term(10, "x")], 10)
    """

    terms: Dict[TermKey, Term]
    constant: float

    def __init__(
        self,
        terms: Optional[Union[Dict[TermKey, Term], List[Term]]] = None,
        constant: Optional[float] = None,
    ) -> None:
        self.constant = 0
        self.terms = {}

        if constant is not None:
            # += b
            self.constant = constant
        if terms is not None:
            # Allows to give two different syntax in the constructor:
            #   - List[Term] is natural
            #   - Dict[str, Term] is useful when constructing a linear expression from the terms of another expression
            if isinstance(terms, dict):
                for term_key, term in terms.items():
                    if not term.is_zero():
                        self.terms[term_key] = term
            elif isinstance(terms, list):
                for term in terms:
                    if not term.is_zero():
                        self.terms[generate_key(term)] = term
            else:
                raise TypeError(
                    f"Terms must be either of type Dict[str, Term] or List[Term], whereas {terms} is of type {type(terms)}"
                )

    def is_zero(self) -> bool:
        return len(self.terms) == 0 and is_zero(self.constant)

    def str_for_constant(self) -> str:
        if is_zero(self.constant):
            return ""
        else:
            return "{:+g}".format(self.constant)

    def __repr__(self) -> str:
        # Useful for debugging tests
        result = ""
        if self.is_zero():
            result += "0"
        else:
            for term in self.terms.values():
                result += repr(term)

            result += self.str_for_constant()

        return result

    def __eq__(self, rhs: object) -> bool:
        return (
            isinstance(rhs, LinearExpression)
            and is_close_abs(self.constant, rhs.constant, EPS)
            and self.terms
            == rhs.terms  # /!\ There may be float equality comparison in the terms values
        )

    def __iadd__(self, rhs: "LinearExpression") -> "LinearExpression":
        if not isinstance(rhs, LinearExpression):
            return NotImplemented
        self.constant += rhs.constant
        aggregated_terms = _merge_dicts(self.terms, rhs.terms, _add_terms)
        self.terms = aggregated_terms
        self.remove_zeros_from_terms()
        return self

    def __add__(self, rhs: "LinearExpression") -> "LinearExpression":
        result = LinearExpression()
        result += self
        result += rhs
        return result

    def __isub__(self, rhs: "LinearExpression") -> "LinearExpression":
        if not isinstance(rhs, LinearExpression):
            return NotImplemented
        self.constant -= rhs.constant
        aggregated_terms = _merge_dicts(self.terms, rhs.terms, _substract_terms)
        self.terms = aggregated_terms
        self.remove_zeros_from_terms()
        return self

    def __sub__(self, rhs: "LinearExpression") -> "LinearExpression":
        result = LinearExpression()
        result += self
        result -= rhs
        return result

    def __neg__(self) -> "LinearExpression":
        result = LinearExpression()
        result -= self
        return result

    def __imul__(self, rhs: "LinearExpression") -> "LinearExpression":
        if not isinstance(rhs, LinearExpression):
            return NotImplemented

        if self.terms and rhs.terms:
            raise ValueError("Cannot multiply two non constant expression")
        else:
            if self.terms:
                left_expr = self
                const_expr = rhs
            else:
                # It is possible that both expr are constant
                left_expr = rhs
                const_expr = self
            if is_close_abs(const_expr.constant, 0, EPS):
                return LinearExpression()
            elif is_close_abs(const_expr.constant, 1, EPS):
                _copy_expression(left_expr, self)
            else:
                left_expr.constant *= const_expr.constant
                for term_key, term in left_expr.terms.items():
                    left_expr.terms[term_key] = Term(
                        term.coefficient * const_expr.constant,
                        term.component_id,
                        term.variable_name,
                        term.time_index,
                        term.scenario_index,
                    )
                _copy_expression(left_expr, self)
        return self

    def __mul__(self, rhs: "LinearExpression") -> "LinearExpression":
        result = LinearExpression()
        result += self
        result *= rhs
        return result

    def __itruediv__(self, rhs: "LinearExpression") -> "LinearExpression":
        if not isinstance(rhs, LinearExpression):
            return NotImplemented

        if rhs.terms:
            raise ValueError("Cannot divide by a non constant expression")
        else:
            if is_close_abs(rhs.constant, 0, EPS):
                raise ZeroDivisionError("Cannot divide expression by zero")
            elif is_close_abs(rhs.constant, 1, EPS):
                return self
            else:
                self.constant /= rhs.constant
                for term_key, term in self.terms.items():
                    self.terms[term_key] = Term(
                        term.coefficient / rhs.constant,
                        term.component_id,
                        term.variable_name,
                        term.time_index,
                        term.scenario_index,
                    )
        return self

    def __truediv__(self, rhs: "LinearExpression") -> "LinearExpression":
        result = LinearExpression()
        result += self
        result /= rhs

        return result

    def remove_zeros_from_terms(self) -> None:
        # TODO: Not optimized, checks could be done directly when doing operations on self.linear_term to avoid copies
        for term_key, term in self.terms.copy().items():
            if is_close_abs(term.coefficient, 0, EPS):
                del self.terms[term_key]


def _copy_expression(src: LinearExpression, dst: LinearExpression) -> None:
    dst.terms = src.terms
    dst.constant = src.constant



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Describes probability distributions used in the models
"""

from abc import ABC
from dataclasses import dataclass
from typing import List

import numpy as np

from gems.expression.expression import ExpressionNode


class AbstractProbabilityLaw(ABC):
    def get_sample(self, size: int) -> List[float]:
        return NotImplemented


@dataclass(frozen=True)
class Normal(AbstractProbabilityLaw):
    mean: ExpressionNode
    standard_deviation: ExpressionNode

    def get_sample(self, size: int) -> List[float]:
        return NotImplemented


@dataclass(frozen=True)
class Uniform(AbstractProbabilityLaw):
    lower_bound: ExpressionNode
    upper_bound: ExpressionNode

    def get_sample(self, size: int) -> List[float]:
        return NotImplemented


@dataclass(frozen=True)
class UniformIntegers(AbstractProbabilityLaw):
    lower_bound: ExpressionNode
    upper_bound: ExpressionNode

    def get_sample(self, size: int) -> List[float]:
        return NotImplemented



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from .benders_decomposed import (
    BendersDecomposedProblem,
    build_benders_decomposed_problem,
)
from .decision_tree import DecisionTreeNode, InterDecisionTimeScenarioConfig
from .optimization import BlockBorderManagement, OptimizationProblem, build_problem
from .output_values import BendersSolution, OutputValues
from .runner import BendersRunner, MergeMPSRunner
from .strategy import MergedProblemStrategy, ModelSelectionStrategy
from .time_block import TimeBlock



from typing import List, Optional
from pydantic import Field
from ..modified_base_model import ModifiedBaseModel
from .gems_component import GemsComponent
from .gems_port_connection import GemsPortConnection
from .gems_area_connection import GemsAreaConnection


class GemsSystem(ModifiedBaseModel):
    id: Optional[str] = None
    model_libraries: Optional[str] = None  # Parsed but unused for n
    components: List[GemsComponent] = Field(default_factory=list)
    connections: Optional[List[GemsPortConnection]] = None
    area_connections: Optional[List[GemsAreaConnection]] = None
    nodes: Optional[List[GemsComponent]] = []



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

TEMPLATE_CLUSTER_TYPE_TO_GET_METHOD = {
    "thermal": "get_thermals",
    "renewable": "get_renewables",
    "st_storage": "get_st_storages",
}
TEMPLATE_CLUSTER_TYPE_TO_DELETE_METHOD = {
    "thermal": "delete_thermal_cluster",
    "renewable": "delete_renewable_cluster",
    "st_storage": "delete_st_storage",
}
STUDY_LEVEL_DELETION = {
    "area": "delete_area",
    "link": "delete_link",
    "binding_constraint": "delete_binding_constraint",
}
STUDY_LEVEL_GET = {
    "area": "get_areas",
    "link": "get_links",
    "binding_constraint": "get_binding_constraints",
}
MATRIX_TYPES_TO_SET_METHOD = {
    "load": "set_load",
    "solar": "set_solar",
    "wind": "set_wind",
}


MATRIX_TYPES_TO_GET_METHOD = {
    "load": "get_load_matrix",
    "solar": "get_solar_matrix",
    "wind": "get_wind_matrix",
}
TIMESERIES_NAME_TO_METHOD = {
    "capacity_direct": "get_capacity_direct",
    "capacity_indirect": "get_capacity_indirect",
    "links_parameters": "get_parameters",
    "thermal_series": "get_series_matrix",
    "thermal_co2": "get_co2_cost_matrix",
    "thermal_fuel": "get_fuel_cost_matrix",
    "renewable_series": "get_timeseries",
    "pmax_injection": "get_pmax_injection",
    "pmax_withdrawal": "get_pmax_withdrawal",
    "lower_rule_curve": "get_lower_rule_curve",
    "upper_rule_curve": "get_upper_rule_curve",
    "storage_inflows": "get_storage_inflows",
    "cost_injection": "get_cost_injection",
    "cost_withdrawal": "get_cost_withdrawal",
    "cost_level": "get_cost_level",
    "cost_variation_injection": "get_cost_variation_injection",
    "cost_variation_withdrawal": "get_cost_variation_withdrawal",
}
TEMPLATE_CLUSTER_TYPE_TO_CLUSTER_PATH = {
    "thermal": "thermal",
    "renewable": "renewables",
    "st_storage": "st-storage",
}
TEMPLATE_LINK_TO_TIMESERIES_FILE_TYPE = {
    "links_parameters": "LINKS_PARAMETERS",
    "capacity_direct": "LINKS_CAPACITIES_DIRECT",
    "capacity_indirect": "LINKS_CAPACITIES_INDIRECT",
}
TEMPLATE_TO_TIMESERIES_FILE_TYPE = {
    "pmax_injection": "ST_STORAGE_PMAX_INJECTION",
    "pmax_withdrawal": "ST_STORAGE_PMAX_WITHDRAWAL",
    "lower_rule_curve": "ST_STORAGE_LOWER_RULE_CURVE",
    "upper_rule_curve": "ST_STORAGE_UPPER_RULE_CURVE",
    "storage_inflows": "ST_STORAGE_INFLOWS",
    "cost_injection": "ST_STORAGE_COST_INJECTION",
    "cost_withdrawal": "ST_STORAGE_COST_WITHDRAWAL",
    "cost_level": "ST_STORAGE_COST_LEVEL",
    "cost_variation_injection": "ST_STORAGE_COST_VARIATION_INJECTION",
    "cost_variation_withdrawal": "ST_STORAGE_COST_VARIATION_WITHDRAWAL",
    "thermal_series": "THERMAL_SERIES",
    "series_co2_cost": "THERMAL_CO2",
    "series_fuel_cost": "THERMAL_FUEL",
    "renewable_series": "RENEWABLE_SERIES",
}

MODEL_NAME_TO_FILE_NAME = {
    "battery": "battery.yaml",
    "link": "link.yaml",
    "load": "load.yaml",
    "solar": "solar.yaml",
    "short-term-storage": "st-storage.yaml",
    "thermal": "thermal.yaml",
    "wind": "wind.yaml",
}

MATRIX_TYPES = ["load", "solar", "wind"]
CLUSTER_TYPES = ["thermal", "renewable", "st_storage"]
LINK_TYPES = ["link"]
AREA_TYPES = ["area"]



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass
from typing import Dict

from gems.expression.expression import (
    AllTimeSumNode,
    ComponentParameterNode,
    ComponentVariableNode,
    ExpressionNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
)

from .expression import (
    AdditionNode,
    Comparator,
    ComparisonNode,
    DivisionNode,
    LiteralNode,
    MultiplicationNode,
    NegationNode,
    ParameterNode,
    ScenarioOperatorNode,
    VariableNode,
)
from .visitor import ExpressionVisitor, visit

_COMPARISON_OPERATOR_TO_STRING: Dict[Comparator, str] = {
    Comparator.LESS_THAN: "<=",
    Comparator.EQUAL: "==",
    Comparator.GREATER_THAN: ">=",
}


@dataclass(frozen=True)
class PrinterVisitor(ExpressionVisitor[str]):
    """
    Produces a string representing the mathematical expression.

    TODO: remove parenthis where not necessary.
    """

    def literal(self, node: LiteralNode) -> str:
        return str(node.value)

    def negation(self, node: NegationNode) -> str:
        return f"-({visit(node.operand, self)})"

    def addition(self, node: AdditionNode) -> str:
        if len(node.operands) == 0:
            return ""
        res = visit(node.operands[0], self)
        for o in node.operands[1:]:
            if isinstance(o, NegationNode):
                res += f" - {visit(o.operand, self)}"
            else:
                res += f" + {visit(o, self)}"
        return f"({res})"

    def multiplication(self, node: MultiplicationNode) -> str:
        left_value = visit(node.left, self)
        right_value = visit(node.right, self)
        return f"({left_value} * {right_value})"

    def division(self, node: DivisionNode) -> str:
        left_value = visit(node.left, self)
        right_value = visit(node.right, self)
        return f"({left_value} / {right_value})"

    def comparison(self, node: ComparisonNode) -> str:
        op = _COMPARISON_OPERATOR_TO_STRING[node.comparator]
        left_value = visit(node.left, self)
        right_value = visit(node.right, self)
        return f"{left_value} {op} {right_value}"

    def variable(self, node: VariableNode) -> str:
        return node.name

    def parameter(self, node: ParameterNode) -> str:
        return node.name

    def comp_variable(self, node: ComponentVariableNode) -> str:
        return f"{node.component_id}.{node.name}"

    def comp_parameter(self, node: ComponentParameterNode) -> str:
        return f"{node.component_id}.{node.name}"

    def pb_variable(self, node: ProblemVariableNode) -> str:
        # TODO
        return f"{node.component_id}.{node.name}"

    def pb_parameter(self, node: ProblemParameterNode) -> str:
        # TODO
        return f"{node.component_id}.{node.name}"

    def time_shift(self, node: TimeShiftNode) -> str:
        return f"({visit(node.operand, self)}.shift({visit(node.time_shift, self)}))"

    def time_eval(self, node: TimeEvalNode) -> str:
        return f"({visit(node.operand, self)}.eval({visit(node.eval_time, self)}))"

    def time_sum(self, node: TimeSumNode) -> str:
        return f"({visit(node.operand, self)}.time_sum({visit(node.from_time, self)}, {visit(node.to_time, self)}))"

    def all_time_sum(self, node: AllTimeSumNode) -> str:
        return f"({visit(node.operand, self)}.time_sum())"

    def scenario_operator(self, node: ScenarioOperatorNode) -> str:
        return f"({visit(node.operand, self)}.{str(node.name)})"

    def port_field(self, node: PortFieldNode) -> str:
        return f"{node.port_name}.{node.field_name}"

    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> str:
        return f"({visit(node.operand, self)}.{node.aggregator})"


def print_expr(expression: ExpressionNode) -> str:
    return visit(expression, PrinterVisitor())



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

import math
from dataclasses import dataclass
from typing import Optional

from gems.expression import (
    AdditionNode,
    ComparisonNode,
    DivisionNode,
    ExpressionNode,
    LiteralNode,
    MultiplicationNode,
    NegationNode,
    ParameterNode,
    VariableNode,
)
from gems.expression.expression import (
    AllTimeSumNode,
    BinaryOperatorNode,
    ComponentParameterNode,
    ComponentVariableNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    ScenarioOperatorNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
)


@dataclass(frozen=True)
class EqualityVisitor:
    abs_tol: float = 0
    rel_tol: float = 0

    def __post_init__(self) -> None:
        if self.abs_tol < 0:
            raise ValueError(
                f"Absolute comparison tolerance must be >= 0, got {self.abs_tol}"
            )
        if self.rel_tol < 0:
            raise ValueError(
                f"Relative comparison tolerance must be >= 0, got {self.rel_tol}"
            )

    def visit(self, left: ExpressionNode, right: ExpressionNode) -> bool:
        if left.__class__ != right.__class__:
            return False
        if isinstance(left, LiteralNode) and isinstance(right, LiteralNode):
            return self.literal(left, right)
        if isinstance(left, NegationNode) and isinstance(right, NegationNode):
            return self.negation(left, right)
        if isinstance(left, AdditionNode) and isinstance(right, AdditionNode):
            return self.addition(left, right)
        if isinstance(left, DivisionNode) and isinstance(right, DivisionNode):
            return self.division(left, right)
        if isinstance(left, MultiplicationNode) and isinstance(
            right, MultiplicationNode
        ):
            return self.multiplication(left, right)
        if isinstance(left, ComparisonNode) and isinstance(right, ComparisonNode):
            return self.comparison(left, right)
        if isinstance(left, VariableNode) and isinstance(right, VariableNode):
            return self.variable(left, right)
        if isinstance(left, ParameterNode) and isinstance(right, ParameterNode):
            return self.parameter(left, right)
        if isinstance(left, ComponentVariableNode) and isinstance(
            right, ComponentVariableNode
        ):
            return self.comp_variable(left, right)
        if isinstance(left, ComponentParameterNode) and isinstance(
            right, ComponentParameterNode
        ):
            return self.comp_parameter(left, right)
        if isinstance(left, ProblemVariableNode) and isinstance(
            right, ProblemVariableNode
        ):
            return self.problem_variable(left, right)
        if isinstance(left, ProblemParameterNode) and isinstance(
            right, ProblemParameterNode
        ):
            return self.problem_parameter(left, right)
        if isinstance(left, TimeShiftNode) and isinstance(right, TimeShiftNode):
            return self.time_shift(left, right)
        if isinstance(left, TimeEvalNode) and isinstance(right, TimeEvalNode):
            return self.time_eval(left, right)
        if isinstance(left, TimeSumNode) and isinstance(right, TimeSumNode):
            return self.time_sum(left, right)
        if isinstance(left, AllTimeSumNode) and isinstance(right, AllTimeSumNode):
            return self.all_time_sum(left, right)
        if isinstance(left, ScenarioOperatorNode) and isinstance(
            right, ScenarioOperatorNode
        ):
            return self.scenario_operator(left, right)
        if isinstance(left, PortFieldNode) and isinstance(right, PortFieldNode):
            return self.port_field(left, right)
        if isinstance(left, PortFieldAggregatorNode) and isinstance(
            right, PortFieldAggregatorNode
        ):
            return self.port_field_aggregator(left, right)
        raise NotImplementedError(f"Equality not implemented for {left.__class__}")

    def literal(self, left: LiteralNode, right: LiteralNode) -> bool:
        return math.isclose(
            left.value, right.value, abs_tol=self.abs_tol, rel_tol=self.rel_tol
        )

    def _visit_operands(
        self, left: BinaryOperatorNode, right: BinaryOperatorNode
    ) -> bool:
        return self.visit(left.left, right.left) and self.visit(left.right, right.right)

    def negation(self, left: NegationNode, right: NegationNode) -> bool:
        return self.visit(left.operand, right.operand)

    def addition(self, left: AdditionNode, right: AdditionNode) -> bool:
        left_ops = left.operands
        right_ops = right.operands
        return len(left_ops) == len(right_ops) and all(
            self.visit(l, r) for l, r in zip(left_ops, right_ops)
        )

    def multiplication(
        self, left: MultiplicationNode, right: MultiplicationNode
    ) -> bool:
        return self._visit_operands(left, right)

    def division(self, left: DivisionNode, right: DivisionNode) -> bool:
        return self._visit_operands(left, right)

    def comparison(self, left: ComparisonNode, right: ComparisonNode) -> bool:
        return left.comparator == right.comparator and self._visit_operands(left, right)

    def variable(self, left: VariableNode, right: VariableNode) -> bool:
        return left.name == right.name

    def parameter(self, left: ParameterNode, right: ParameterNode) -> bool:
        return left.name == right.name

    def comp_variable(
        self, left: ComponentVariableNode, right: ComponentVariableNode
    ) -> bool:
        return left.name == right.name and left.component_id == right.component_id

    def comp_parameter(
        self, left: ComponentParameterNode, right: ComponentParameterNode
    ) -> bool:
        return left.name == right.name and left.component_id == right.component_id

    def problem_variable(
        self, left: ProblemVariableNode, right: ProblemVariableNode
    ) -> bool:
        return (
            left.name == right.name
            and left.component_id == right.component_id
            and left.time_index == right.time_index
            and left.scenario_index == right.scenario_index
        )

    def problem_parameter(
        self, left: ProblemParameterNode, right: ProblemParameterNode
    ) -> bool:
        return (
            left.name == right.name
            and left.component_id == right.component_id
            and left.time_index == right.time_index
            and left.scenario_index == right.scenario_index
        )

    def time_shift(self, left: TimeShiftNode, right: TimeShiftNode) -> bool:
        return self.visit(left.time_shift, right.time_shift) and self.visit(
            left.operand, right.operand
        )

    def time_eval(self, left: TimeEvalNode, right: TimeEvalNode) -> bool:
        return self.visit(left.eval_time, right.eval_time) and self.visit(
            left.operand, right.operand
        )

    def time_sum(self, left: TimeSumNode, right: TimeSumNode) -> bool:
        return (
            self.visit(left.from_time, right.from_time)
            and self.visit(left.to_time, right.to_time)
            and self.visit(left.operand, right.operand)
        )

    def all_time_sum(self, left: AllTimeSumNode, right: AllTimeSumNode) -> bool:
        return self.visit(left.operand, right.operand)

    def scenario_operator(
        self, left: ScenarioOperatorNode, right: ScenarioOperatorNode
    ) -> bool:
        return left.name == right.name and self.visit(left.operand, right.operand)

    def port_field(self, left: PortFieldNode, right: PortFieldNode) -> bool:
        return left.port_name == right.port_name and left.field_name == right.field_name

    def port_field_aggregator(
        self, left: PortFieldAggregatorNode, right: PortFieldAggregatorNode
    ) -> bool:
        return left.aggregator == right.aggregator and self.visit(
            left.operand, right.operand
        )


def expressions_equal(
    left: ExpressionNode, right: ExpressionNode, abs_tol: float = 0, rel_tol: float = 0
) -> bool:
    """
    True if both expression nodes are equal. Literal values may be compared with absolute or relative tolerance.
    """
    return EqualityVisitor(abs_tol, rel_tol).visit(left, right)


def expressions_equal_if_present(
    lhs: Optional[ExpressionNode], rhs: Optional[ExpressionNode]
) -> bool:
    if lhs is None and rhs is None:
        return True
    elif lhs is None or rhs is None:
        return False
    else:
        return expressions_equal(lhs, rhs)



from typing import Optional, Union
from ..modified_base_model import ModifiedBaseModel


class GemsComponentParameter(ModifiedBaseModel):
    id: str
    time_dependent: bool = False
    scenario_dependent: bool = False
    value: Union[float, str]
    scenario_group: Optional[str] = None



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
import logging
import shutil
from pathlib import Path
from typing import Optional, Union

import pandas as pd
from antares.craft.exceptions.exceptions import ReferencedObjectDeletionNotAllowed
from antares.craft.model.link import Link
from antares.craft.model.study import Study, read_study_local

from gems.input_converter.src.config import (
    LINK_TYPES,
    MATRIX_TYPES,
    MATRIX_TYPES_TO_SET_METHOD,
    MODEL_NAME_TO_FILE_NAME,
    STUDY_LEVEL_DELETION,
    STUDY_LEVEL_GET,
    TEMPLATE_CLUSTER_TYPE_TO_DELETE_METHOD,
    TEMPLATE_CLUSTER_TYPE_TO_GET_METHOD,
)
from gems.input_converter.src.data_preprocessing.data_classes import ConversionMode
from gems.input_converter.src.data_preprocessing.preprocessing import (
    ModelConversionPreprocessor,
)
from gems.input_converter.src.data_preprocessing.thermal import ThermalDataPreprocessing
from gems.input_converter.src.parsing import (
    ConversionTemplate,
    ObjectProperties,
    VirtualObjectsRepository,
    parse_conversion_template,
)
from gems.input_converter.src.utils import dump_to_yaml, read_yaml_file, resolve_path
from gems.study.parsing import (
    InputAreaConnections,
    InputComponent,
    InputComponentParameter,
    InputPortConnections,
    InputSystem,
)

ANTARES_HISTORIC_LIB_ID = "antares-historic"
MODEL_TEMPLATE_FOLDER = Path(__file__).parents[1] / "data" / "model_configuration"
LIBS_FOLDER = "model-libraries"
SERIES_FOLDER = "data-series"

# TODO: Move all global variables in a config class, that is used in AntaresStudyConverter constructor


class AntaresStudyConverter:
    def __init__(
        self,
        study_input: Union[Path, Study],
        logger: logging.Logger,
        mode: str = "full",
        output_folder: Path = Path("/tmp/"),
        period: Optional[int] = None,
        lib_paths: Optional[list[str]] = None,
        models_to_convert: list[str] = list(MODEL_NAME_TO_FILE_NAME.keys()),
        modeler_scenario_builder_file: Optional[Path] = None,
    ):
        """
        Initialize processor
        """
        self.logger = logger
        self.period: int = period if period else 168
        self.lib_paths: list[str] = lib_paths if lib_paths else []
        self.models_to_convert = models_to_convert
        self.modeler_scenario_builder_file = modeler_scenario_builder_file
        try:
            self.mode = ConversionMode(mode)
        except ValueError:
            raise ValueError(
                f"Invalid conversion mode: {mode}, possible values are {[conv_mode.value for conv_mode in ConversionMode]}"
            )

        # TODO: The logic is still too complicated, needs more refacto / to understand why thermal preprocessing sometimes needs different paths
        study_input_path = (
            study_input.path.stem
            if isinstance(study_input, Study)
            else study_input.stem
        )
        self.output_folder = output_folder / study_input_path

        if self.mode == ConversionMode.HYBRID:
            # In hybrid mode, the output is the input study from which we replace converted components by Gems ones, hence we copy the original study
            shutil.copytree(
                study_input.path if isinstance(study_input, Study) else study_input,
                self.output_folder,
                dirs_exist_ok=True,
            )
            if isinstance(study_input, Path):
                study_input = self.output_folder
        else:
            # In full mode, the output is a full Gems study so no need to copy the original study, we start "from scratch"
            self.output_folder.mkdir(parents=True, exist_ok=True)

        if isinstance(study_input, Study):
            # We have a different way of managing thermal preprocessing files, because in this case we want to modify the study_path.
            # But in the same moment we dont want the preprocessing files in the modified study_path
            study_input.path = self.output_folder
            self.study = study_input
        else:
            # TODO: Check whether this dinstinction is needed
            if mode == ConversionMode.HYBRID:
                self.study = read_study_local(resolve_path(self.output_folder))
            else:
                self.study = read_study_local(resolve_path(study_input))
        self.output_system_path = self.output_folder / "input" / "system.yml"
        self.areas = self.study.get_areas()
        self.legacy_objects: list[ObjectProperties] = []

    def _convert_thermal_to_component_list(
        self,
        lib_id: str,
        virtual_objects: VirtualObjectsRepository,
        components: list,
        connections: list,
        area_connections: list,
        scenario_group: Optional[str] = None,
    ) -> tuple[list[InputComponent], list[InputPortConnections]]:
        self.logger.info("Converting thermals to component list...")
        # Add thermal components for each area
        for area in self.areas.values():
            if area.id not in virtual_objects.areas:
                resolved_virtual_objects = virtual_objects.resolve_template(
                    "${area}", area.id
                )
                thermals = area.get_thermals()
                for thermal in thermals.values():
                    if thermal.id not in resolved_virtual_objects.thermals:
                        tdp = ThermalDataPreprocessing(thermal, self.output_folder)
                        components.append(
                            InputComponent(
                                id=f"{thermal.area_id}_{thermal.id}",
                                model=f"{lib_id}.thermal",
                                parameters=[
                                    tdp.generate_component_parameter("p_min_cluster"),
                                    tdp.generate_component_parameter("nb_units_min"),
                                    tdp.generate_component_parameter("nb_units_max"),
                                    tdp.generate_component_parameter(
                                        "nb_units_max_variation_forward", self.period
                                    ),
                                    tdp.generate_component_parameter(
                                        "nb_units_max_variation_backward", self.period
                                    ),
                                    InputComponentParameter(
                                        id="unit_count",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.unit_count,
                                    ),
                                    InputComponentParameter(
                                        id="p_min_unit",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.min_stable_power,
                                    ),
                                    InputComponentParameter(
                                        id="efficiency",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.efficiency,
                                    ),
                                    InputComponentParameter(
                                        id="p_max_unit",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.nominal_capacity,
                                    ),
                                    InputComponentParameter(
                                        id="generation_cost",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.marginal_cost,
                                    ),
                                    InputComponentParameter(
                                        id="fixed_cost",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.fixed_cost,
                                    ),
                                    InputComponentParameter(
                                        id="startup_cost",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.startup_cost,
                                    ),
                                    InputComponentParameter(
                                        id="d_min_up",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.min_up_time,
                                    ),
                                    InputComponentParameter(
                                        id="d_min_down",
                                        time_dependent=False,
                                        scenario_dependent=False,
                                        value=thermal.properties.min_down_time,
                                    ),
                                    tdp.generate_component_parameter("p_max_cluster"),
                                ],
                                scenario_group=scenario_group,
                            )
                        )
                        if self.mode == ConversionMode.FULL:
                            connections.append(
                                InputPortConnections(
                                    component1=f"{thermal.area_id}_{thermal.id}",
                                    port1="balance_port",
                                    component2=f"{thermal.area_id}",
                                    port2="balance_port",
                                )
                            )
                        else:
                            area_connections.append(
                                InputAreaConnections(
                                    component=f"{thermal.area_id}_{thermal.id}",
                                    port="balance_port",
                                    area=f"{thermal.area_id}",
                                )
                            )
                            self.legacy_objects.append(
                                ObjectProperties(
                                    type="thermal",
                                    area=thermal.area_id,
                                    link=None,
                                    cluster=thermal.id,
                                    binding_constraint_id=None,
                                    field=None,
                                )
                            )
        return components, connections

    def _convert_area_to_component_list(
        self, lib_id: str, excluded_areas: Optional[list[str]] = None
    ) -> list[InputComponent]:
        components = []
        self.logger.info("Converting areas to component list...")
        for area in self.areas.values():
            if not excluded_areas or area.id not in excluded_areas:
                components.append(
                    InputComponent(
                        id=area.id,
                        model=f"{lib_id}.area",
                        parameters=[
                            InputComponentParameter(
                                id="ens_cost",
                                time_dependent=False,
                                scenario_dependent=False,
                                value=area.properties.energy_cost_unsupplied,
                            ),
                            InputComponentParameter(
                                id="spillage_cost",
                                time_dependent=False,
                                scenario_dependent=False,
                                value=area.properties.energy_cost_spilled,
                            ),
                        ],
                    )
                )
        return components

    def _delete_legacy_objects(self) -> None:
        for legacy_component in self.legacy_objects:
            try:
                if legacy_component.type in STUDY_LEVEL_DELETION:
                    if legacy_component.type == "area":
                        id = legacy_component.area
                    elif legacy_component.type == "link":
                        id = legacy_component.link
                    elif legacy_component.type == "binding_constraint":
                        id = legacy_component.binding_constraint_id
                    else:
                        # Should not happen
                        pass
                    getattr(self.study, STUDY_LEVEL_DELETION[legacy_component.type])(
                        getattr(self.study, STUDY_LEVEL_GET[legacy_component.type])()[
                            id
                        ]
                    )
                elif (
                    legacy_component.type in TEMPLATE_CLUSTER_TYPE_TO_DELETE_METHOD
                    and legacy_component.area is not None
                ):
                    getattr(
                        self.areas[legacy_component.area],
                        TEMPLATE_CLUSTER_TYPE_TO_DELETE_METHOD[legacy_component.type],
                    )(
                        getattr(
                            self.areas[legacy_component.area],
                            TEMPLATE_CLUSTER_TYPE_TO_GET_METHOD[legacy_component.type],
                        )()[legacy_component.cluster]
                    )
                elif (
                    legacy_component.type in MATRIX_TYPES_TO_SET_METHOD
                    and legacy_component.area is not None
                ):
                    # To "delete" legacy wind, solar or load object, we simply set an empty timeseries
                    getattr(
                        self.areas[legacy_component.area],
                        MATRIX_TYPES_TO_SET_METHOD[legacy_component.type],
                    )(pd.DataFrame())

            except ReferencedObjectDeletionNotAllowed:
                self.logger.warning(
                    f"Item {legacy_component} will not be deleted because it is referenced in a binding constraint"
                )
            except NotImplementedError:
                self.logger.warning(
                    f"Failure to delete {legacy_component} because the method is not implemented yet on antares craft"
                )

        self.legacy_objects = []

    def _iterate_through_model(
        self,
        resolved_conversion_template: ConversionTemplate,
        components: list,
        connections: list,
        area_connections: list,
        mp: ModelConversionPreprocessor,
    ) -> None:
        parameters = [
            InputComponentParameter(
                id=param.id,
                time_dependent=bool(param.time_dependent),
                scenario_dependent=bool(param.scenario_dependent),
                value=mp.convert_param_value(param.id, param.value),
            )
            for param in resolved_conversion_template.component.parameters
        ]
        scenario_group = getattr(resolved_conversion_template, "scenario_group", None)
        kwargs = {}
        if scenario_group is not None:
            kwargs["scenario_group"] = scenario_group

        components.append(
            InputComponent(
                id=(resolved_conversion_template.component.id).replace(" ", "_"),
                model=resolved_conversion_template.model,
                parameters=parameters,
                **kwargs,
            )
        )

        if self.mode == ConversionMode.HYBRID:
            for area_connection in resolved_conversion_template.area_connections:
                # TODO: Improve logic
                if "." in area_connection.component:
                    component_parts = area_connection.component.split(".")
                    component_value = getattr(
                        self.study.get_links()[component_parts[0]], component_parts[1]
                    )
                else:
                    component_value = area_connection.component
                component_value = component_value.replace(" ", "_")
                if "." in area_connection.area:
                    area_parts = area_connection.area.split(".")
                    area_value = getattr(
                        self.study.get_links()[area_parts[0]], area_parts[1]
                    )
                else:
                    area_value = area_connection.area
                area_value = area_value.replace(" ", "_")
                area_connections.append(
                    InputAreaConnections(
                        component=component_value,
                        port=area_connection.port,
                        area=area_value,
                    )
                )
            # TODO : Simplify usage, use directly legacy_objectis_to_delete
            for item in resolved_conversion_template.legacy_objects_to_delete:
                self.legacy_objects.append(item.object_properties)
        else:
            for connection in resolved_conversion_template.connections:
                # TODO: Factorize logic with previous connections
                treated_components = []
                for component in [connection.component1, connection.component2]:
                    if "." in component:
                        component_parts = component.split(".")
                        component_value = getattr(
                            self.study.get_links()[component_parts[0]],
                            component_parts[1],
                        )
                    else:
                        component_value = component
                    treated_components.append(component_value.replace(" ", "_"))

                connections.append(
                    InputPortConnections(
                        component1=treated_components[0],
                        port1=connection.port1,
                        component2=treated_components[1],
                        port2=connection.port2,
                    )
                )

    def _convert_model_to_component_list(
        self,
        conversion_template: ConversionTemplate,
        virtual_objects: VirtualObjectsRepository = VirtualObjectsRepository(),
    ) -> tuple[
        list[InputComponent], list[InputPortConnections], list[InputAreaConnections]
    ]:
        components: list[InputComponent] = []
        connections: list[InputPortConnections] = []
        area_connections: list[InputAreaConnections] = []

        model_area_pattern = f"${{{conversion_template.template_parameters[0].name}}}"

        model_preprocessor = ModelConversionPreprocessor(
            self.study, self.mode, self.output_folder
        )

        try:
            if conversion_template.name in LINK_TYPES:
                for link in self.study.get_links().values():
                    if not self.is_virtual_link(link, virtual_objects):
                        resolved_template = conversion_template.resolve_template(
                            model_area_pattern, link.id
                        )
                        self._iterate_through_model(
                            resolved_template,
                            components,
                            connections,
                            area_connections,
                            model_preprocessor,
                        )
            else:
                if conversion_template.name == "thermal":
                    # Legacy conversion for thermal cluster
                    self._convert_thermal_to_component_list(
                        self.get_model_name_among_libs("thermal"),
                        virtual_objects,
                        components,
                        connections,
                        area_connections,
                        scenario_group=getattr(
                            conversion_template, "scenario_group", None
                        ),
                    )
                    return components, connections, area_connections
                for area in self.areas.values():
                    if area.id not in virtual_objects.areas:
                        resolved_template = conversion_template.resolve_template(
                            model_area_pattern, area.id
                        )
                        cluster_type = next(
                            (
                                template.cluster_type
                                for template in conversion_template.template_parameters
                            ),
                            None,
                        )
                        if cluster_type:
                            for cluster_id in getattr(
                                area, TEMPLATE_CLUSTER_TYPE_TO_GET_METHOD[cluster_type]
                            )():
                                # We have already resolved areas, now need to resolve cluster ids
                                resolved_template = resolved_template.resolve_template(
                                    f"${{{cluster_type}}}", cluster_id
                                )
                                self._iterate_through_model(
                                    resolved_template,
                                    components,
                                    connections,
                                    area_connections,
                                    model_preprocessor,
                                )

                        elif conversion_template.name in MATRIX_TYPES:
                            if all(
                                model_preprocessor.check_timeseries_validity(
                                    param.value
                                )
                                for param in resolved_template.component.parameters
                            ):
                                self._iterate_through_model(
                                    resolved_template,
                                    components,
                                    connections,
                                    area_connections,
                                    model_preprocessor,
                                )
                        else:
                            self._iterate_through_model(
                                resolved_template,
                                components,
                                connections,
                                area_connections,
                                model_preprocessor,
                            )
        except (KeyError, FileNotFoundError) as e:
            self.logger.error(
                f"Error while converting model to component list: {e}. "
                "Please check the model configuration file."
            )
            return components, connections, area_connections

        return components, connections, area_connections

    @staticmethod
    def is_virtual_link(link: Link, virtual_objects: VirtualObjectsRepository) -> bool:
        return (
            link.id in virtual_objects.links
            or link.area_from_id in virtual_objects.areas
            or link.area_to_id in virtual_objects.areas
        )

    @staticmethod
    def _extract_lib_and_model_ids(path: str) -> tuple[str, list]:
        lib_data = read_yaml_file(Path(path))["library"]
        models = lib_data.get("models", [])
        return lib_data["id"], [model["id"] for model in models]

    def _check_converted_models_are_in_libs(
        self, model_conversion_templates: dict[str, ConversionTemplate]
    ) -> None:
        lib_to_model_ids = {}
        for lib_path in self.lib_paths:
            lib_id, model_ids = self._extract_lib_and_model_ids(lib_path)
            lib_to_model_ids[lib_id] = model_ids

        for model in self.models_to_convert:
            lib_id, model_id = model_conversion_templates[model].model.split(".")
            if lib_id not in lib_to_model_ids:
                raise ValueError(
                    f"Library {lib_id} has not been found in provided libraries"
                )
            if model_id not in lib_to_model_ids[lib_id]:
                raise ValueError(
                    f"Model {model_id} has not been found in library {lib_id}"
                )

    # TODO: Does not depend on self for now, but will be once the config is a class attribute
    def _get_model_conversion_template(self, model: str) -> ConversionTemplate:
        model_conversion_template_file = (
            MODEL_TEMPLATE_FOLDER / MODEL_NAME_TO_FILE_NAME[model]
        )
        if not model_conversion_template_file.exists():
            raise FileNotFoundError(
                f"The model configuration file for {model} has not been found at the location {model_conversion_template_file}"
            )
        # TODO: Parse yaml file with Pydantic to return a proper Python object rather than a generic dict
        with model_conversion_template_file.open() as template:
            return parse_conversion_template(template)

    def _copy_libs_to_model_librairies(self) -> None:
        # Retrieve library files and put it in the output study (as fro now libs must be contained in modeler studies)
        dest_dir = self.output_folder / "input" / LIBS_FOLDER
        dest_dir.mkdir(parents=True, exist_ok=True)
        for path in self.lib_paths:
            shutil.copy2(path, dest_dir)

    def _create_dataseries_dir(self) -> None:
        dest_dir = self.output_folder / "input" / SERIES_FOLDER
        dest_dir.mkdir(parents=True, exist_ok=True)

    def get_model_name_among_libs(self, model_name: str) -> str:
        for lib_path in self.lib_paths:
            lib_name, file_model = self._extract_lib_and_model_ids(lib_path)
            if model_name in file_model:
                return lib_name
        return ANTARES_HISTORIC_LIB_ID

    def _convert_single_model(
        self,
        conversion_template: ConversionTemplate,
        virtual_objects: VirtualObjectsRepository,
        components: list[InputComponent],
        connections: list[InputPortConnections],
        area_connections: list[InputAreaConnections],
    ) -> None:
        self.logger.info(
            f"Converting components of model {conversion_template.name}..."
        )

        (
            components_from_model,
            connections_from_model,
            area_connections_from_model,
        ) = self._convert_model_to_component_list(conversion_template, virtual_objects)

        components.extend(components_from_model)
        connections.extend(connections_from_model)
        area_connections.extend(area_connections_from_model)

    def convert_study_to_input_system(self) -> InputSystem:
        self._copy_libs_to_model_librairies()
        self._copy_scenario_builder()
        self._create_dataseries_dir()
        model_conversion_templates = self._build_model_conversion_templates()
        self._check_converted_models_are_in_libs(model_conversion_templates)
        virtual_objects = self._build_virtual_objects_repo(model_conversion_templates)
        components: list[InputComponent] = []
        connections: list[InputPortConnections] = []
        area_connections: list[InputAreaConnections] = []
        for model in self.models_to_convert:
            conversion_template = model_conversion_templates[model]
            self._convert_single_model(
                conversion_template,
                virtual_objects,
                components,
                connections,
                area_connections,
            )
        if self.mode == ConversionMode.HYBRID:
            self._delete_legacy_objects()
        else:
            components.extend(
                self._convert_area_to_component_list(
                    ANTARES_HISTORIC_LIB_ID, virtual_objects.areas
                )
            )
        system = InputSystem(
            id=self.study.name,
            components=components,
            connections=connections or None,
            area_connections=area_connections or None,
        )
        data = system.model_dump(exclude_none=True)
        return InputSystem(**data)

    def _build_model_conversion_templates(self) -> dict[str, ConversionTemplate]:
        model_conversion_templates: dict[str, ConversionTemplate] = {}
        for model in self.models_to_convert:
            model_conversion_templates[model] = self._get_model_conversion_template(
                model
            )
        return model_conversion_templates

    def _build_virtual_objects_repo(
        self, model_conversion_templates: dict[str, ConversionTemplate]
    ) -> VirtualObjectsRepository:
        virtual_objects = VirtualObjectsRepository()
        for model in self.models_to_convert:
            virtual_objects_this_model = model_conversion_templates[
                model
            ].get_excluded_objects_ids()
            virtual_objects.add(virtual_objects_this_model)
        return virtual_objects

    def process_all(self) -> None:
        system = self.convert_study_to_input_system()
        self.logger.info("Dumping input system into yaml file...")
        dump_to_yaml(model=system, output_path=self.output_system_path)

    def _copy_scenario_builder(self) -> None:
        if not self.modeler_scenario_builder_file:
            return

        dest = self.output_folder / "input" / "data-series"
        dest.mkdir(parents=True, exist_ok=True)

        dest_file = dest / "modeler-scenariobuilder.dat"  # enforce name

        try:
            shutil.copy2(self.modeler_scenario_builder_file, dest_file)
            self.logger.info(f"Copied scenario builder file to {dest_file}")
        except Exception as e:
            self.logger.warning(f"Failed to copy scenario builder file: {e}")



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from .common import ProblemContext, ValueType
from .constraint import Constraint
from .model import Model, ModelPort, model
from .parameter import Parameter, float_parameter, int_parameter
from .port import PortField, PortFieldDefinition, PortFieldId, PortType
from .variable import Variable, float_variable, int_variable



from pathlib import Path
from typing import Any, Union

import pandas as pd
from antares.craft.model.binding_constraint import BindingConstraint, ConstraintTerm
from antares.craft.model.link import Link
from antares.craft.model.study import Study

from gems.input_converter.src.config import (
    MATRIX_TYPES_TO_GET_METHOD,
    TEMPLATE_CLUSTER_TYPE_TO_GET_METHOD,
    TIMESERIES_NAME_TO_METHOD,
)
from gems.input_converter.src.data_preprocessing.data_classes import ConversionMode
from gems.input_converter.src.parsing import ConversionValue
from gems.input_converter.src.utils import check_dataframe_validity, save_to_file

ALLOWED_TYPES: list = [
    "binding_constraint",
    "thermal",
    "link",
    "st_storage",
    "load",
    "solar",
    "wind",
]
SERIES_FOLDER = "data-series"


class ModelConversionPreprocessor:
    preprocessed_values: dict[str, float] = {}
    param_id: str

    def __init__(self, study: Study, mode: ConversionMode, output_folder: Path):
        self.study = study
        self.mode = mode
        self.output_folder = output_folder
        self.output_file = Path(".")
        self.file_path = Path(".")

    def calculate_matrix_data_values(
        self, obj: ConversionValue, type_resource: str
    ) -> pd.DataFrame:
        if not obj.object_properties or not obj.object_properties.area:
            raise ValueError(
                f"Object properties and its area from {obj} must not be None"
            )
        area: str = obj.object_properties.area
        self.file_path = Path(f"{self.param_id}_{area}.tsv")
        self.output_file = self.output_folder / "input" / SERIES_FOLDER / self.file_path
        return getattr(
            self.study.get_areas()[area], MATRIX_TYPES_TO_GET_METHOD[type_resource]
        )()

    def calculate_link_data_values(self, obj: ConversionValue) -> pd.DataFrame:
        if (
            not obj.object_properties
            or not obj.object_properties.link
            or not obj.object_properties.field
        ):
            raise ValueError(
                f"Object properties, its link, and field from {obj} must not be None"
            )
        link_id = obj.object_properties.link

        link: Link = self.study.get_links()[link_id]
        self.file_path = Path(
            f"{self.param_id}_{link.area_from_id}_{link.area_to_id}.tsv"
        )
        self.output_file = self.output_folder / "input" / SERIES_FOLDER / self.file_path
        return getattr(link, TIMESERIES_NAME_TO_METHOD[obj.object_properties.field])()

    def calculate_cluster_data_values(
        self, type_resource: str, obj: ConversionValue
    ) -> Union[Any, pd.DataFrame]:
        if (
            not obj.object_properties
            or not obj.object_properties.area
            or not obj.object_properties.field
        ):
            raise ValueError(
                f"Object properties, its area, and field from {obj} must not be None"
            )
        area: str = obj.object_properties.area
        if area not in self.study.get_areas():
            raise KeyError(f"Area {area} is not found in the study")
        cluster = getattr(
            self.study.get_areas()[area],
            TEMPLATE_CLUSTER_TYPE_TO_GET_METHOD[type_resource],
        )()[obj.object_properties.cluster]
        if obj.object_properties.field in TIMESERIES_NAME_TO_METHOD:
            time_series = getattr(
                cluster, TIMESERIES_NAME_TO_METHOD[obj.object_properties.field]
            )()
        else:
            cluster_properties = getattr(cluster, "properties")
            field_name = obj.object_properties.field
            value = getattr(cluster_properties, field_name)
            if type_resource == "thermal":
                self.preprocessed_values[self.param_id] = value
            return value
        self.file_path = Path(
            f"{self.param_id}_{area}_{obj.object_properties.cluster}.tsv"
        )
        self.output_file = self.output_folder / "input" / SERIES_FOLDER / self.file_path
        return time_series

    def calculate_binding_constraint_data_values(
        self, obj: ConversionValue
    ) -> Union[float, pd.Series, pd.DataFrame]:
        if (
            not obj.object_properties
            or not obj.object_properties.field
            or not obj.object_properties.binding_constraint_id
        ):
            raise ValueError(
                f"Object properties, its field, and binding constraint ID from {obj} must not be None"
            )
        binding: BindingConstraint = self.study.get_binding_constraints()[
            obj.object_properties.binding_constraint_id
        ]
        term: ConstraintTerm = binding.get_terms()[obj.object_properties.field]
        if obj.operation:
            return obj.operation.execute(term.weight)
        else:
            return term.weight

    def calculate_value(self, obj: ConversionValue) -> Union[float, str]:
        if obj.object_properties is None or obj.object_properties.type is None:
            raise ValueError(f"Object properties {obj} must not be None")
        type_resource: str = obj.object_properties.type
        time_series: pd.DataFrame = pd.DataFrame()
        if type_resource in ["load", "wind", "solar"]:
            time_series = self.calculate_matrix_data_values(obj, type_resource)
            save_to_file(time_series, self.output_file)
            return str(self.file_path).removesuffix(".tsv")
        elif type_resource == "binding_constraint":
            # TODO No timeseries linked to binding constraints for the moment
            return self.calculate_binding_constraint_data_values(obj)  # type: ignore
        elif type_resource == "link":
            time_series = self.calculate_link_data_values(obj)
        elif type_resource in ["st_storage", "thermal", "renewable"]:
            data = self.calculate_cluster_data_values(type_resource, obj)
            if isinstance(data, pd.DataFrame):
                time_series = data
            else:
                return data

        if getattr(obj, "column", None) is not None:
            time_series: pd.Series = time_series.iloc[:, obj.column]  # type: ignore

            if getattr(obj, "operation") and obj.operation is not None:
                parameter_value: Union[
                    pd.Series, pd.DataFrame, float
                ] = obj.operation.execute(time_series, self.preprocessed_values)
                if isinstance(parameter_value, float):
                    self.preprocessed_values[self.param_id] = parameter_value
                    return parameter_value
                if isinstance(parameter_value, pd.Series):
                    save_to_file(parameter_value, self.output_file)
            else:
                save_to_file(time_series, self.output_file)
        else:
            save_to_file(time_series, self.output_file)

        return str(self.file_path).removesuffix(".tsv")

    def convert_param_value(
        self, id: str, value_content: ConversionValue
    ) -> Union[str, float]:
        self.param_id = id
        if value_content.constant is not None:
            return value_content.constant

        if not value_content.object_properties:
            raise ValueError(f"Object properties from {value_content} must not be None")
        if value_content.object_properties.type not in ALLOWED_TYPES:
            raise ValueError(
                f"Unknown value type: {value_content.object_properties.type}"
            )
        return self.calculate_value(value_content)

    def check_timeseries_validity(self, value_content: ConversionValue) -> bool:
        if value_content.constant is not None:
            return True

        if (
            not value_content.object_properties
            or not value_content.object_properties.area
            or not value_content.object_properties.type
        ):
            raise ValueError(
                f"Object properties, its area, and type from {value_content} must not be None"
            )
        # TODO : Do not duplicate logic with the one in convert_param_value
        if value_content.object_properties.type not in ALLOWED_TYPES:
            raise ValueError(
                f"Unknown value type: {value_content.object_properties.type}"
            )
        time_series: pd.DataFrame = getattr(
            self.study.get_areas()[value_content.object_properties.area],
            MATRIX_TYPES_TO_GET_METHOD[value_content.object_properties.type],
        )()
        return check_dataframe_validity(time_series)



from pydantic import BaseModel


class ModifiedBaseModel(BaseModel):
    class Config:
        alias_generator = lambda snake: snake.replace("_", "-")
        extra = "forbid"
        populate_by_name = True



from dataclasses import dataclass
import pandas as pd


@dataclass
class PyPSAComponentData:
    pypsa_model_id: str
    constant_data: pd.DataFrame
    time_dependent_data: dict[str, pd.DataFrame]
    gems_model_id: str
    pypsa_params_to_gems_params: dict[str, str]
    pypsa_params_to_gems_connections: dict[str, tuple[str, str]]

    def check_params_consistency(self) -> None:
        for key in self.pypsa_params_to_gems_params:
            self._check_key_in_constant_data(key)
        for key in self.pypsa_params_to_gems_connections:
            self._check_key_in_constant_data(key)

    def _check_key_in_constant_data(self, key: str) -> None:
        if key not in self.constant_data.columns:
            raise ValueError(
                f"Parameter {key} not available in constant data, defining all available parameters for model {self.pypsa_model_id}"
            )
